# GAP code to compute the normal forms for matrices in GL(n, Z/p^2Z) or (Z/p^2Z)^(n x n)
# under GL(n, Z/p^2Z). 
#
# To use this, call
# ConjugacyClassesGLZmodp2Z(n, p);
# or
# NormalFormsZmodp2Z(n, p);
# with the appropriate n and p.
#
# Copyright by Sebastian Jambor, 2011
# sebastian.jambor@rwth-aachen.de





# calculate the number of irreducible polynomials of degree n in F_p[x]
NrIrreduciblePols := function(p, n)
    local fcts, res, f, i, j;

    fcts := Collected(Factors(n));
    res := p^n;

    for f in fcts do
        res := res - p^(n/f[1]);
    od;

    for i in [1..Size(fcts)] do
        for j in [i+1..Size(fcts)] do
            res := res + p^((n/fcts[i][1])/fcts[j][1]);
        od;
    od;

    return res/n;
end;


# calculate all irreducible polynomials of degree n in F_p[x]
IrreduciblePols := function(p, n)
    local nr, res, F, x, mp;
    

    nr := NrIrreduciblePols;
    res := [];
    F := GF(p^n);

    for x in F do
        mp := MinimalPolynomial(GF(p), x);
        if Degree(mp) = n and not mp in res then
            Add(res, mp);
        fi;

        if Size(res) = nr then
            break;
        fi;
    od;


    return res;
end;


# given a list L = [L_1, ..., L_k] of natural numbers and an index
# i between 1 and \sum L, let 1 <= j <= k be the number such that
# \sum_{l = 1}^{j-1} L_l < i <= \sum_{l = 1}^j L_l.
#
# Return the pair [j, i-\sum_{l = 1}^{j-1} L_l]
GetRelativeIndices := function(L, i) 
    local n, j;

    n := L[1];

    j := 1;

    while i > n do
        j := j+1;
        n := n + L[j];
    od;

    return [j, i - n + L[j]];
end;


# given a block diagonal matrix with blocks B, return the
# entry (i,j) of this matrix.
GetEntry := function(B, i, j)
    local L, reli, relj;

    L := List(B, Size);
    reli := GetRelativeIndices(L, i);
    relj := GetRelativeIndices(L, j);

    if reli[1] <> relj[1] then
        return 0;
    fi;

    return B[reli[1]][reli[2]][relj[2]];
end;

BlockDiagonalMatrix := function(B)
    local n;

    # remove empty blocks
    B := Filtered(B, b -> Size(b) > 0);

    n := Sum(List(B, Size));

    return List([1..n], i -> List([1..n], j -> GetEntry(B, i, j)))*One(B[1][1][1]); # multiply by one, so the zeroes become zeroes of the field
end;


# compute the Size of the centralizer algebra
CentOrder := function(pi, q)
    local res, p1, p2;

    res := 1;

    # now compute the size of the Jacobson radical
    for p1 in pi do
        for p2 in pi do
            res := res*q^Minimum(p1, p2);
        od;
    od;

    return res;
end;

# compute the order of the centralizer unit group
CentUnitsOrder := function(pi, q)
    local gl, res, coll, c, p1, p2;
    
    # compute the order of GL(d, q)
    gl := function(d, q)
        local res, i;

        res := 1;
        for i in [0..d-1] do
            res := res*(q^d - q^i);
        od;

        return res;
    end;

    res := 1;
    coll := Collected(pi);

    # compute the order of the direct product of the GL's sitting on top
    for c in coll do
        res := res*gl(c[2], q);
    od;

    # now compute the size of the Jacobson radical
    for p1 in pi do
        for p2 in pi do
            if p1 = p2 then
                res := res*q^(p1 - 1);
            else
                res := res*q^Minimum(p1, p2);
            fi;
        od;
    od;

    return res;
end;

# return all Frobenius normal forms of matrices with characteristic polynomial mu^k
AllFrobeniusForms := function(q, mu, k)
    local part, res, pi;

    part := Partitions(k);

    res := [];
    
    for pi in part do
        Add(res, [BlockDiagonalMatrix(List(pi, i -> CompanionMat(mu^i))), CentUnitsOrder(pi, q^Degree(mu))]);
    od;

    return res;
end;


# return all representatives of conjugacy classes of matrices of dimension n \times n,
# where the divisors of the minimal polynomial are in the list pols, together with the
# order of their centralizers.
# Thus: Return a list of pairs, where the first entry in each pair
# is a representative of a conjugacy class, and the second entry is
# the order of the centralizer of this particular representative.
#
# most efficient if pols is sorted descreasingly
#
# Idea: First construct the hauptspace for pols[1], then recursively the complement.
AllRepresentatives := function(n, q, pols)
    local mu, d, m, res, k, frob, rep, f, r;

    if Size(pols) = 0 then
        return [];
    fi;

    mu := pols[1];
    d := Degree(mu);
    m := Int(n/d);

    res := [];
    for k in [1..m] do
        # construct the hauptspace for mu
        frob := AllFrobeniusForms(q, mu, k);
        
        if d*k = n then
            # there is no room for a complement.
            Append(res, frob);
        else
            # get all possibilities for complements, and patch them together
            rep := AllRepresentatives(n - k*d, q, pols{[2..Size(pols)]});

            for f in frob do
                for r in rep do
                    Add(res, [BlockDiagonalMatrix([f[1], r[1]]), f[2]*r[2]]);
                od;
            od;
        fi;
    od;

    if Size(pols) > 1 then
        # add possibilities, where there is no hauptspace for mu.
        Append(res, AllRepresentatives(n, q, pols{[2..Size(pols)]}));
    fi;

    return res;
end;

AllPols := function(p, n)
    local k, res;

    res := [];
    
    k := n;
    while k > 0 do
        Append(res, IrreduciblePols(p, k));
        k := k - 1;
    od;

    return res;
end;

AllNonzeroPols := function(p, n)
    local k, res, tmp;

    res := [];
    
    k := n;
    while k > 0 do
        tmp := IrreduciblePols(p, k);

        if k = 1 then
            Remove(tmp, Position(tmp, Indeterminate(GF(p))));
        fi;

        Append(res, tmp);

        k := k - 1;
    od;

    return res;
end;


# convert a matrix over GF(p) or Z to a maxrix over Z/p^2Z
modpSquared := function(m)
    local p;
    p := Characteristic(m);

    return List(m, r -> List(r, e -> Int(e)))*One(Integers mod p^2);
end;

modpSquaredPol := function(mu)
    local p, c, x;

    p := Characteristic(mu);

    c := CoefficientsOfUnivariatePolynomial(mu);
    c := List(c, e -> Int(e))*One(Integers mod p^2);

    x := Indeterminate(Integers mod p^2);

    return Sum(List([1..Size(c)], i -> c[i]*x^(i-1)));
end;


# return all companion matrices in (Z/p^2Z)^{n \times n}, where
# the polynomial is in the residue class of mu mod pZ/p^2Z[x],
AllCompanionMatrices := function(mu)
    local tmp, i, x, res, c, n, p;

    n := LeadingMonomial(mu)[2];
    p := Factors(Characteristic(mu))[1];

    tmp := [];
    for i in [1..n] do
        Add(tmp, [1..p]);
    od;

    x := Indeterminate(Integers mod p^2);

    res := [];

    for c in Cartesian(tmp) do
        Add(res, CompanionMat(mu + Sum(List([1..n], i -> p*c[i]*x^(i-1)))));
    od;

    return res;
end;


# converts a matrix m in GL(k, GF(p^d)) to a matrix in GL(d*k, GF(p)), where q = p^d
ExpandMat := function(m, q)
    local p, F, B, k, d, g, C, i, j, coeffvectors, matrices, blocks;

    p := Characteristic(m);

    F := GF(q);
    B := Basis(F);

    k := Size(m);
    d := DegreeFFE(Z(q));

    g := CompanionMat(MinimalPolynomial(GF(p), Z(q)));

    C := [IdentityMat(d)];

    for i in [1..d-1] do
        Add(C, g^i);
    od;

    coeffvectors := List(m, r -> List(r, e -> Coefficients(B, e)));
    matrices := List(coeffvectors, r -> List(r, c -> Sum(List([1..d], i -> c[i]*C[i]))));

    blocks := [];
    for i in [1..k] do
        for j in [1..k] do
            Add(blocks, [i,j,matrices[i][j]]);
        od;
    od;

    return MatrixByBlockMatrix(BlockMatrix(blocks, k, k));
end;

# If m = n and J(a) is the Jordan block of dimension n to the eigenvalue a, 
# this method returns J(a)^exp.
# If n > m, then this method returns a mxn-matrix, where J(a)^exp is in
# the left mxm submatrix.
# If m > n, then return an mxn-matrix, where J(a)^exp is in the lower
# nxn submatrix
JordanBlock := function(m, n, a, exp)
    local tmp;
    
    tmp := function(i,j)
        local k;

        if m > n then
            i := i - (m - n);
        fi;

        k := i - j;
        if k = 0 then
            return a^(exp - k);
        elif 1 <= k and k < exp then
            return exp*a^(exp - k);
        elif k = exp then
            return One(a);
        else
            return Zero(a);
        fi;
    end;

    return List([1..m], i -> List([1..n], j -> tmp(i,j)));
end;



# rowdims and coldims are lists of positive integers, giving the dimensions of the
# block matrices. blocks are given as in the standard gap BlockMatrix function, i.e.,
# a list, whose elements are of the form [i, j, m], where i,j are indices which
# describe the position of the block, and m is the actual block (so it has to be a 
# matrix of dimension rowdims[i]xcoldims[j]. If the block at position (i,j) is not 
# in the list blocks, then every entry of this block is set to default.
#
# In contrast to the usual gap BlockMatrix function, the returned matrix is a usual
# matrix, not a sparse matrix, and arbitary row and column dimensions for the blocks
# are allowed.
MyBlockMatrix := function(rowdims, coldims, blocks, default)
    local getBlockIndex, getEntry, m, n;

    # If the block (i,j) is defined in the list of blocks, return the index.
    # Otherwise return -1.
    getBlockIndex := function(i,j)
        local k;
        for k in [1..Size(blocks)] do
            if blocks[k][1] = i and blocks[k][2] = j then
                return k;
            fi;
        od;
        return -1;
    end;
    
    getEntry := function(i,j)
        local r, c, b;

        r := GetRelativeIndices(rowdims, i);
        c := GetRelativeIndices(coldims, j);

        b := getBlockIndex(r[1], c[1]);

        if b = -1 then
            return default;
        fi;

        return blocks[b][3][r[2]][c[2]];
    end;

    m := Sum(rowdims);
    n := Sum(coldims);

    return List([1..m], i -> List([1..n], j -> getEntry(i,j)));
end;


# Construct the centralizer algebra of the Jordan matrix with
# eigenvalue a and partition pi
CentralizerAlgebra := function(a, pi)
    local basis, i, j, exp;

    basis := [];

    for i in [1..Size(pi)] do
        for j in [1..Size(pi)] do
            for exp in [0..Minimum(pi[i], pi[j])-1] do
#                Add(basis, MyBlockMatrix(pi, pi, [[i,j,JordanBlock(pi[i], pi[j], a, exp)]], Zero(a)));
                Add(basis, MyBlockMatrix(pi, pi, [[i,j,JordanBlock(pi[i], pi[j], Zero(a), exp)]], Zero(a)));
            od;
        od;
    od;

    return VectorSpace(Field(a), basis);
end;

# Construct the unit group of the centralizer algebra of the Jordan matrix with
# eigenvalue a and partition pi
CentralizerUnitGroup := function(a, pi)
    local generators, collectBlocks, blocks, n, startdim, enddim, b, g, gens, blowupg, E, v, i, j, c, r;


    # like the Collected function, but doesn't sort the output, but the
    # elements in the output appear in the same order as in the input
    collectBlocks := function(pi)
        local result, tmp, count, i;

        result := [];
        tmp := pi[1];
        count := 1;

        for i in [2..Size(pi)] do
            if pi[i] = tmp then
                count := count + 1;
            else
                Add(result, [tmp, count]);
                tmp := pi[i];
                count := 1;
            fi;
        od;
        Add(result, [tmp, count]);

        return result;
    end;

    generators := [];

    blocks := collectBlocks(pi);
    n := Sum(pi);
    startdim := 0;
    enddim := n;

    for b in blocks do
        # first, get generators for the GL's on the diagonal
        gens := GeneratorsOfGroup(GL(b[2], Field(a)));
        enddim := enddim - b[1]*b[2];
        for g in gens do
            if g <> One(g) then
                # avoid the identity element, which occurs in generators for GL(1, F)
                blowupg := BlockDiagonalMatrix([
                        IdentityMat(startdim, Field(a)), 
                        KroneckerProduct(g, IdentityMat(b[1], Field(a))), 
                        IdentityMat(enddim, Field(a))
                    ]);
                Add(generators, blowupg);
            fi;
        od;

        # next, add generators for the vector spaces under the GL
        E := List([1..b[2]], i -> List([1..b[2]], j -> Zero(a)));
        E[1][1] := One(a);
        for exp in [1..b[1]-1] do
            for v in Basis(Field(a)) do
                Add(generators, BlockDiagonalMatrix([
                        IdentityMat(startdim, Field(a)),
#                        IdentityMat(b[1]*b[2], Field(a)) + v*KroneckerProduct(E, JordanBlock(b[1], b[1], a, exp)),
                        IdentityMat(b[1]*b[2], Field(a)) + v*KroneckerProduct(E, JordanBlock(b[1], b[1], Zero(a), exp)),
                        IdentityMat(enddim, Field(a))
                    ]));
            od;
        od;

        startdim := startdim + b[1]*b[2];
    od;

    # finally, add generators to compute the vector spaces on the off-diagonal blocks
    r := 1;
    for i in [1..Size(blocks)] do
        c := Sum(List(blocks{[1..i]}, k -> k[2])) + 1;
        for j in [i+1..Size(blocks)] do
#            Add(generators, IdentityMat(n, Field(a)) + MyBlockMatrix(pi, pi, [[r,c,JordanBlock(pi[r], pi[c], a, 0)]], Zero(a)));
#            Add(generators, IdentityMat(n, Field(a)) + MyBlockMatrix(pi, pi, [[c,r,JordanBlock(pi[c], pi[r], a, 0)]], Zero(a)));
            Add(generators, IdentityMat(n, Field(a)) + MyBlockMatrix(pi, pi, [[r,c,JordanBlock(pi[r], pi[c], Zero(a), 0)]], Zero(a)));
            Add(generators, IdentityMat(n, Field(a)) + MyBlockMatrix(pi, pi, [[c,r,JordanBlock(pi[c], pi[r], Zero(a), 0)]], Zero(a)));

            c := c + blocks[j][2];
        od;

        r := r + blocks[i][2];
    od;
    
    return Group(generators);
end;


# Let J be the Jordan matrix with eigenvalue a and partition pi.
# Then this method constructs representatives of a basis
# of the cokernel of the map (X -> Jx - XJ).
CokernelBasis := function(a, pi)
    local basis, delta, n, i, j, rowoffset, coloffset, k, d;

    # construct a function in two parameters k,l, which returns 1 if and only if k = i and j = l
    # and zero otherwise
    delta := function(i,j)
        return function(k,l)
            if i = k and j = l then
                return One(a);
            else
                return Zero(a);
            fi;
        end;
    end;

    basis := [];

    n := Sum(pi);

    for i in [1..Size(pi)] do
        for j in [1..Size(pi)] do
            rowoffset := Sum(pi{[1..i-1]});
            coloffset := Sum(pi{[1..j]});

            for k in [0..Minimum(pi[i], pi[j])-1] do
                d := delta(rowoffset + 1, coloffset - k);
                Add(basis, List([1..n], a -> List([1..n], b -> d(a,b))));
            od;
        od;
    od;

    return basis;
end;



# auxillary function to compute the row vector in F^n
# with 1 at position i.
UnitVector := function(i, n, F)
    local result;

    result := List([1..n], j -> 0);
    result[i] := 1;
    return result*One(F);
end;



# return all representatives of conjugacy classes in GL(n, Z/p^2Z),
# where the divisors of the minimal polynomial mod p are in the list pols,
# and p is the characteristic of the ring of pols, together with the
# order of their centralizers.
# Thus: Return a list of pairs, where the first entry in each pair
# is a representative of a conjugacy class, and the second entry is
# the order of the centralizer of this particular representative.
#
# most efficient if pols is sorted descreasingly
#
# Idea: First construct the hauptspace for pols[1], then recursively the complement.
AllRepresentativesMod := function(n, pols)
    local mu, d, m, res, k, repk, rep, f, r, R, pi, part, p, base, lift, additional, centorder, A, G, orb, o, kernelbasis, cokernelbasis, cokerToKer, kerToCoker, J, repcoeffs, gens, gensmat, Gtrans, baseChanges, s, B, b, F, v, i, baseChangeMatrix;

    if Size(pols) = 0 then
        return [];
    fi;

    p := Characteristic(pols[1]);

    mu := pols[1];
    d := Degree(mu);
    m := Int(n/d);

    res := [];
    for k in [1..m] do
        # If we are at the last polynomial in the list, there can not be any complements,
        # so the current block should be the last. Don't bother to compute blocks which
        # are too small.
        if Size(pols) = 1 and k*d < n then
            continue;
        fi;
        
        # construct the hauptspace for mu
        part := Partitions(k);

        repk := [];
        for pi in part do
            centorder := CentOrder(pi, p^d);
            if pi[1] = k then
                # cyclic case, i.e., the minimal polynomial is the characteristic polynomial
                # The centralizer in this case is an extension of (\F_p[x]/<mu^k>, +) by
                # (\F_p[x]/<mu^k>)
                centorder := centorder*p^(d*(k-1))*(p^d - 1);
                Append(repk, List(AllCompanionMatrices(modpSquaredPol(mu^k)), m -> [m, centorder]));
            elif pi[1] = 1 then
                # semi-simple case
                lift := modpSquaredPol(mu);
                base := BlockDiagonalMatrix(List(pi, i -> CompanionMat(lift)));
                additional := AllRepresentatives(k, p^d, AllPols(p^d, k));
                additional := List(additional, a -> [modpSquared(ExpandMat(a[1], p^d)), a[2]]);

                Append(repk, List(additional, a -> [base + p*a[1], centorder*a[2]]));
            else
                # general case
                # Print("General case for mu = ", mu, ", pi = ", pi, "\n");
                A := CentralizerAlgebra(Z(p^d), pi);
                G := CentralizerUnitGroup(Z(p^d), pi);

                # we are interested in the dual action of the conjugation action,
                # so compute the orbits by manually constructing the inverse transpose action.
                kernelbasis := Basis(A);
                gens := GeneratorsOfGroup(G);;

                gensmat := List(gens, g -> List(kernelbasis, b -> Coefficients(kernelbasis, g^(-1)*b*g)));

                Gtrans := Group(List(gensmat, TransposedMat));

                orb := OrbitsDomain(Gtrans, GF(p^d)^Dimension(A));


                # We now have the orbits of G on the dual of A.
                # What we really want are the orbits of G on the cokernel.
                cokernelbasis := CokernelBasis(Z(p^d), pi);

                cokerToKer := List(cokernelbasis, c -> List(kernelbasis, b -> Trace(c*b)));
                kerToCoker := cokerToKer^(-1);

                # at the moment, the matrix and the orbit representatives are
                # in Jordan form, tensored up. Convert this to Frobenius form.
                # Note that every Jordan block is cyclic, so we can easily
                # compute a base consisting of cyclic vectors and their images.
                #
                # We have to transpose the Jordan blocks, so that the first unit
                # vector is always cyclic (GAP uses rows). After that, transpose
                # the base change matrix.
                baseChanges := [];
                for s in pi do
                    B := TransposedMat(ExpandMat(JordanBlock(s,s,RootsOfUPol(GF(p^d), mu)[1],1), p^d));
                    v := UnitVector(1, Size(B), GF(p));
                    base := [v];
                    for i in [2..Size(B)] do
                        Add(base, base[i-1]*B);
                    od;

                    Add(baseChanges, TransposedMat(base));
                od;
                baseChangeMatrix := BlockDiagonalMatrix(baseChanges);

                J := ExpandMat(BlockDiagonalMatrix(List(pi, s -> JordanBlock(s,s,RootsOfUPol(GF(p^d), mu)[1],1))), p^d);
                F := modpSquared(baseChangeMatrix^(-1)*J*baseChangeMatrix);

                for o in orb do
                    repcoeffs := Representative(o)*kerToCoker;
                    rep := F + p*modpSquared(baseChangeMatrix^(-1)*ExpandMat(Sum(List([1..Size(repcoeffs)], i -> repcoeffs[i]*cokernelbasis[i])), p^d)*baseChangeMatrix);

                    Add(repk, [rep, centorder*Order(G)/Size(o)]);
                od;

                # Print("centorder = ", centorder, "\n");
            fi;
        od;

        if d*k = n then
            # there is no room for a complement.
            Append(res, repk);
        else
            # get all possibilities for complements, and patch them together
            rep := AllRepresentativesMod(n - k*d, pols{[2..Size(pols)]});

            for f in repk do
                for r in rep do
                    Add(res, [BlockDiagonalMatrix([f[1], r[1]]), f[2]*r[2]]);
                od;
            od;
        fi;
    od;

    if Size(pols) > 1 then
        # add possibilities, where there is no hauptspace for mu.
        Append(res, AllRepresentativesMod(n, pols{[2..Size(pols)]}));
    fi;

    return res;
end;


# temporary method!
# sends a matrix in the kernel of delta_a to a matrix in the cokernel of delta_a,
# for partition [2,1,1] and prime p = 2
# for partition [2,1] and prime p = 2
ker2coker := function(m)
#    return [[m[2][1], m[1][1], m[2][3], m[2][4]], [0,0,0,0], [0, m[3][1], m[3][3], m[3][4]], [0, m[4][1], m[4][3], m[4][4]]]*One(GF(2));
    return [[m[2][1], m[1][1], m[2][3]], [0,0,0], [0, m[3][1], m[3][3]]]*One(GF(2));
end;



elementaryMatrix := function(n, i, j)
    local res;

    res := List([1..n], i -> List([1..n], j -> 0));
    res[i][j] := 1;

    return res;
end;



# Compute the conjugacy classes of GL(n, Z/p^2Z).
#
# This returns a list of pairs.
# The first entry of each pair is a representative of a conjugacy class,
# the second entry is the order of its centralizer in GL(n, Z/p^2Z).
ConjugacyClassesGLZmodp2Z := function(n, p)
    local pols;

    pols := AllNonzeroPols(p, n);
    return AllRepresentativesMod(n, pols);
end;


# Compute the normal forms of (Z/p^2Z)^(n x n)
#
# This returns a list of pairs.
# The first entry of each pair is a representative of a conjugacy class,
# the second entry is the order of its centralizer in GL(n, Z/p^2Z).
NormalFormsZmodp2Z := function(n, p)
    local pols;

    pols := AllPols(p, n);
    return AllRepresentativesMod(n, pols);
end;
